@@ -20,6 +20,12 @@ from .mouse_event import MouseEvent
 
 MOUSEMASK = curses.ALL_MOUSE_EVENTS | curses.REPORT_MOUSE_POSITION
 
+# This escape is not available with a capname from terminfo unlike
+# tsl (to_status_line), so it's hardcoded here. It's used just like tsl,
+# but it sets the icon title (WM_ICON_NAME) instead of the window title
+# (WM_NAME).
+ESCAPE_ICON_TITLE = '\033]1;'
+
 _ASCII = ''.join(chr(c) for c in range(32, 127))
 
 
@@ -64,8 +70,11 @@ class UI(  # pylint: disable=too-many-instance-attributes,too-many-public-method
         self.status = None
         self.console = None
         self.pager = None
+        self.multiplexer = None
         self._draw_title = None
         self._tmux_automatic_rename = None
+        self._tmux_title = None
+        self._screen_title = None
         self.browser = None
 
         if fm is not None:
@@ -109,23 +118,14 @@ class UI(  # pylint: disable=too-many-instance-attributes,too-many-public-method
             self.is_set_up = True
             self.setup()
             self.win.addstr("loading...")
-            self.win.refresh()
             self._draw_title = curses.tigetflag('hs')  # has_status_line
 
-            # Save tmux setting `automatic-rename`
-            if self.settings.update_tmux_title and 'TMUX' in os.environ:
-                try:
-                    self._tmux_automatic_rename = check_output(
-                        ['tmux', 'show-window-options', '-v', 'automatic-rename']).strip()
-                except CalledProcessError:
-                    self._tmux_automatic_rename = None
+        self.win.refresh()
 
         self.update_size()
         self.is_on = True
 
-        if self.settings.update_tmux_title and 'TMUX' in os.environ:
-            sys.stdout.write("\033kranger\033\\")
-            sys.stdout.flush()
+        self.handle_multiplexer()
 
         if 'vcsthread' in self.__dict__:
             self.vcsthread.unpause()
@@ -136,13 +136,12 @@ class UI(  # pylint: disable=too-many-instance-attributes,too-many-public-method
             self.vcsthread.pause()
             self.vcsthread.paused.wait()
 
+        if self.fm.image_displayer:
+            self.fm.image_displayer.quit()
+
         self.win.keypad(0)
         curses.nocbreak()
         curses.echo()
-        try:
-            curses.curs_set(1)
-        except curses.error:
-            pass
         if self.settings.mouse_enabled:
             _setup_mouse(dict(value=False))
         curses.endwin()
@@ -171,19 +170,7 @@ class UI(  # pylint: disable=too-many-instance-attributes,too-many-public-method
             del self.__dict__['vcsthread']
         DisplayableContainer.destroy(self)
 
-        # Restore tmux setting `automatic-rename`
-        if self.settings.update_tmux_title and 'TMUX' in os.environ:
-            if self._tmux_automatic_rename:
-                try:
-                    check_output(['tmux', 'set-window-option',
-                                  'automatic-rename', self._tmux_automatic_rename])
-                except CalledProcessError:
-                    pass
-            else:
-                try:
-                    check_output(['tmux', 'set-window-option', '-u', 'automatic-rename'])
-                except CalledProcessError:
-                    pass
+        self.restore_multiplexer_name()
 
         self.suspend()
 
@@ -235,8 +222,10 @@ class UI(  # pylint: disable=too-many-instance-attributes,too-many-public-method
         for key in keys:
             self.handle_key(key)
 
-    def handle_input(self):
+    def handle_input(self):  # pylint: disable=too-many-branches
         key = self.win.getch()
+        if key == curses.KEY_ENTER:
+            key = ord('\n')
         if key == 27 or (key >= 128 and key < 256):
             # Handle special keys like ALT+X or unicode here:
             keys = [key]
@@ -272,6 +261,9 @@ class UI(  # pylint: disable=too-many-instance-attributes,too-many-public-method
                 else:
                     if not self.fm.input_is_blocked():
                         self.handle_key(key)
+            elif key == -1 and not os.isatty(sys.stdin.fileno()):
+                # STDIN has been closed
+                self.fm.exit()
 
     def setup(self):
         """Build up the UI by initializing widgets."""
@@ -376,16 +368,18 @@ class UI(  # pylint: disable=too-many-instance-attributes,too-many-public-method
             try:
                 fixed_cwd = cwd.encode('utf-8', 'surrogateescape'). \
                     decode('utf-8', 'replace')
-                fmt_tup = (
+                escapes = [
                     curses.tigetstr('tsl').decode('latin-1'),
-                    fixed_cwd,
-                    curses.tigetstr('fsl').decode('latin-1'),
-                )
+                    ESCAPE_ICON_TITLE
+                ]
+                bel = curses.tigetstr('fsl').decode('latin-1')
+                fmt_tups = [(e, fixed_cwd, bel) for e in escapes]
             except UnicodeError:
                 pass
             else:
-                sys.stdout.write("%sranger:%s%s" % fmt_tup)
-                sys.stdout.flush()
+                for fmt_tup in fmt_tups:
+                    sys.stdout.write("%sranger:%s%s" % fmt_tup)
+                    sys.stdout.flush()
 
         self.win.refresh()
 
@@ -469,6 +463,62 @@ class UI(  # pylint: disable=too-many-instance-attributes,too-many-public-method
         else:
             self.titlebar.throbber = string
 
+    # Handles window renaming behaviour of the terminal multiplexers
+    # GNU Screen and Tmux
+    def handle_multiplexer(self):
+        if self.settings.update_tmux_title:
+            if 'TMUX' in os.environ:
+                # Stores the automatic-rename setting
+                # prints out a warning if the allow-rename in tmux is not set
+                tmux_allow_rename = check_output(
+                    ['tmux', 'show-window-options', '-v',
+                     'allow-rename']).strip()
+                if tmux_allow_rename == 'off':
+                    self.fm.notify('Warning: allow-rename not set in Tmux!',
+                                   bad=True)
+                elif self._tmux_title is None:
+                    self._tmux_title = check_output(
+                        ['tmux', 'display-message', '-p', '#W']).strip()
+                else:
+                    try:
+                        self._tmux_automatic_rename = check_output(
+                            ['tmux', 'show-window-options', '-v',
+                             'automatic-rename']).strip()
+                        if self._tmux_automatic_rename == 'on':
+                            check_output(['tmux', 'set-window-option',
+                                          'automatic-rename', 'off'])
+                    except CalledProcessError:
+                        pass
+            elif 'screen' in os.environ['TERM'] and self._screen_title is None:
+                # Stores the screen window name before renaming it
+                # gives out a warning if $TERM is not "screen"
+                try:
+                    self._screen_title = check_output(
+                        ['screen', '-Q', 'title']).strip()
+                except CalledProcessError:
+                    self._screen_title = None
+
+            sys.stdout.write("\033kranger\033\\")
+            sys.stdout.flush()
+
+    # Restore window name
+    def restore_multiplexer_name(self):
+        try:
+            if 'TMUX' in os.environ:
+                if self._tmux_automatic_rename:
+                    check_output(['tmux', 'set-window-option',
+                                  'automatic-rename',
+                                  self._tmux_automatic_rename])
+                else:
+                    check_output(['tmux', 'set-window-option', '-u',
+                                  'automatic-rename'])
+                if self._tmux_title:
+                    check_output(['tmux', 'rename-window', self._tmux_title])
+            elif 'screen' in os.environ['TERM'] and self._screen_title:
+                check_output(['screen', '-X', 'title', self._screen_title])
+        except CalledProcessError:
+            self.fm.notify("Could not restore window-name!", bad=True)
+
     def hint(self, text=None):
         self.status.hint = text
 
